@@ -8,12 +8,12 @@
  * modify it under the terms of the GNU General Public License
  * as published by the Free Software Foundation; either version 2
  * of the License, or (at your option) any later version.
-
+ *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
-
+ *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
@@ -32,7 +32,6 @@
 #include "scumm/imuse/imuse_internal.h"
 #include "scumm/imuse/instrument.h"
 #include "scumm/resource.h"
-#include "scumm/saveload.h"
 #include "scumm/scumm.h"
 
 namespace Scumm {
@@ -46,7 +45,7 @@ namespace Scumm {
 IMuseInternal::IMuseInternal() :
 	_native_mt32(false),
 	_enable_gs(false),
-	_sc55(false),
+	_isAmiga(false),
 	_midi_adlib(NULL),
 	_midi_native(NULL),
 	_sysex(NULL),
@@ -159,9 +158,11 @@ bool IMuseInternal::isMT32(int sound) {
 	case MKTAG('S', 'P', 'K', ' '):
 		return false;
 
-	case MKTAG('A', 'M', 'I', ' '):
-	case MKTAG('R', 'O', 'L', ' '):
-		return true;
+	case MKTAG('A', 'M', 'I', ' '): // MI2 Amiga
+		return false;
+
+	case MKTAG('R', 'O', 'L', ' '): // Unfortunately FOA Amiga also uses this resource type
+		return !_isAmiga;
 
 	case MKTAG('M', 'A', 'C', ' '): // Occurs in the Mac version of FOA and MI2
 		return false;
@@ -175,6 +176,9 @@ bool IMuseInternal::isMT32(int sound) {
 			return true;
 		else
 			return false;
+
+	default:
+		break;
 	}
 
 	// Old style 'RO' has equivalent properties to 'ROL'
@@ -201,7 +205,9 @@ bool IMuseInternal::isMIDI(int sound) {
 	case MKTAG('S', 'P', 'K', ' '):
 		return false;
 
-	case MKTAG('A', 'M', 'I', ' '):
+	case MKTAG('A', 'M', 'I', ' '): // Amiga (return true, since the driver is initalized as native midi)
+		return true;
+
 	case MKTAG('R', 'O', 'L', ' '):
 		return true;
 
@@ -211,6 +217,9 @@ bool IMuseInternal::isMIDI(int sound) {
 	case MKTAG('G', 'M', 'D', ' '):
 	case MKTAG('M', 'I', 'D', 'I'): // Occurs in Sam & Max
 		return true;
+
+	default:
+		break;
 	}
 
 	// Old style 'RO' has equivalent properties to 'ROL'
@@ -238,9 +247,11 @@ bool IMuseInternal::supportsPercussion(int sound) {
 	case MKTAG('S', 'P', 'K', ' '):
 		return false;
 
-	case MKTAG('A', 'M', 'I', ' '):
-	case MKTAG('R', 'O', 'L', ' '):
-		return true;
+	case MKTAG('A', 'M', 'I', ' '): // MI2 Amiga
+		return false;
+
+	case MKTAG('R', 'O', 'L', ' '): // Roland LAPC/MT-32/CM32L track, but also used by INDY4 Amiga
+		return !_isAmiga;
 
 	case MKTAG('M', 'A', 'C', ' '): // Occurs in the Mac version of FOA and MI2
 		// This is MIDI, i.e. uses MIDI style program changes, but without a
@@ -250,6 +261,9 @@ bool IMuseInternal::supportsPercussion(int sound) {
 	case MKTAG('G', 'M', 'D', ' '):
 	case MKTAG('M', 'I', 'D', 'I'): // Occurs in Sam & Max
 		return true;
+
+	default:
+		break;
 	}
 
 	// Old style 'RO' has equivalent properties to 'ROL'
@@ -349,7 +363,7 @@ void IMuseInternal::pause(bool paused) {
 	update_volumes();
 	_music_volume = vol;
 
-	// Fix for Bug #817871. The MT-32 apparently fails
+	// Fix for Bug #1263. The MT-32 apparently fails
 	// sometimes to respond to a channel volume message
 	// (or only uses it for subsequent note events).
 	// The result is hanging notes on pause. Reportedly
@@ -363,70 +377,50 @@ void IMuseInternal::pause(bool paused) {
 	_paused = paused;
 }
 
-int IMuseInternal::save_or_load(Serializer *ser, ScummEngine *scumm, bool fixAfterLoad) {
-	Common::StackLock lock(_mutex, "IMuseInternal::save_or_load()");
-	const SaveLoadEntry mainEntries[] = {
-		MKLINE(IMuseInternal, _queue_end, sleUint8, VER(8)),
-		MKLINE(IMuseInternal, _queue_pos, sleUint8, VER(8)),
-		MKLINE(IMuseInternal, _queue_sound, sleUint16, VER(8)),
-		MKLINE(IMuseInternal, _queue_adding, sleByte, VER(8)),
-		MKLINE(IMuseInternal, _queue_marker, sleByte, VER(8)),
-		MKLINE(IMuseInternal, _queue_cleared, sleByte, VER(8)),
-		MKLINE(IMuseInternal, _master_volume, sleByte, VER(8)),
-		MKLINE(IMuseInternal, _trigger_count, sleUint16, VER(8)),
-		MKLINE(IMuseInternal, _snm_trigger_index, sleUint16, VER(54)),
-		MKARRAY(IMuseInternal, _channel_volume[0], sleUint16, 8, VER(8)),
-		MKARRAY(IMuseInternal, _volchan_table[0], sleUint16, 8, VER(8)),
-		MKEND()
-	};
+static void syncWithSerializer(Common::Serializer &s, CommandQueue &cq) {
+	s.syncArray(cq.array, 8, Common::Serializer::Uint16LE, VER(23));
+}
 
-	const SaveLoadEntry cmdQueueEntries[] = {
-		MKARRAY(CommandQueue, array[0], sleUint16, 8, VER(23)),
-		MKEND()
-	};
+static void syncWithSerializer(Common::Serializer &s, ImTrigger &it) {
+	s.syncAsSint16LE(it.sound, VER(54));
+	s.syncAsByte(it.id, VER(54));
+	s.syncAsUint16LE(it.expire, VER(54));
+	s.syncArray(it.command, 8, Common::Serializer::Uint16LE, VER(54));
+}
 
-	// VolumeFader is obsolete.
-	const SaveLoadEntry volumeFaderEntries[] = {
-		MK_OBSOLETE(VolumeFader, player, sleUint16, VER(8), VER(16)),
-		MK_OBSOLETE(VolumeFader, active, sleUint8, VER(8), VER(16)),
-		MK_OBSOLETE(VolumeFader, curvol, sleUint8, VER(8), VER(16)),
-		MK_OBSOLETE(VolumeFader, speed_lo_max, sleUint16, VER(8), VER(16)),
-		MK_OBSOLETE(VolumeFader, num_steps, sleUint16, VER(8), VER(16)),
-		MK_OBSOLETE(VolumeFader, speed_hi, sleInt8, VER(8), VER(16)),
-		MK_OBSOLETE(VolumeFader, direction, sleInt8, VER(8), VER(16)),
-		MK_OBSOLETE(VolumeFader, speed_lo, sleInt8, VER(8), VER(16)),
-		MK_OBSOLETE(VolumeFader, speed_lo_counter, sleUint16, VER(8), VER(16)),
-		MKEND()
-	};
-
-	const SaveLoadEntry snmTriggerEntries[] = {
-		MKLINE(ImTrigger, sound, sleInt16, VER(54)),
-		MKLINE(ImTrigger, id, sleByte, VER(54)),
-		MKLINE(ImTrigger, expire, sleUint16, VER(54)),
-		MKARRAY(ImTrigger, command[0], sleUint16, 8, VER(54)),
-		MKEND()
-	};
+void IMuseInternal::saveLoadIMuse(Common::Serializer &s, ScummEngine *scumm, bool fixAfterLoad) {
+	Common::StackLock lock(_mutex, "IMuseInternal::saveLoadIMuse()");
 
 	int i;
 
-	ser->saveLoadEntries(this, mainEntries);
-	ser->saveLoadArrayOf(_cmd_queue, ARRAYSIZE(_cmd_queue), sizeof(_cmd_queue[0]), cmdQueueEntries);
-	ser->saveLoadArrayOf(_snm_triggers, ARRAYSIZE(_snm_triggers), sizeof(_snm_triggers[0]), snmTriggerEntries);
+	s.syncAsByte(_queue_end, VER(8));
+	s.syncAsByte(_queue_pos, VER(8));
+	s.syncAsUint16LE(_queue_sound, VER(8));
+	s.syncAsByte(_queue_adding, VER(8));
+	s.syncAsByte(_queue_marker, VER(8));
+	s.syncAsByte(_queue_cleared, VER(8));
+	s.syncAsByte(_master_volume, VER(8));
+	s.syncAsUint16LE(_trigger_count, VER(8));
+	s.syncAsUint16LE(_snm_trigger_index, VER(54));
+	s.syncArray(_channel_volume, 8, Common::Serializer::Uint16LE, VER(8));
+	s.syncArray(_volchan_table, 8, Common::Serializer::Uint16LE, VER(8));
+	s.syncArray(_cmd_queue, ARRAYSIZE(_cmd_queue), syncWithSerializer);
+	s.syncArray(_snm_triggers, ARRAYSIZE(_snm_triggers), syncWithSerializer);
 
 	// The players
 	for (i = 0; i < ARRAYSIZE(_players); ++i)
-		_players[i].saveLoadWithSerializer(ser);
+		_players[i].saveLoadWithSerializer(s);
 
 	// The parts
 	for (i = 0; i < ARRAYSIZE(_parts); ++i)
-		_parts[i].saveLoadWithSerializer(ser);
+		_parts[i].saveLoadWithSerializer(s);
 
 	{
 		// Load/save the instrument definitions, which were revamped with V11.
 		Part *part = &_parts[0];
-		if (ser->getVersion() >= VER(11)) {
+		if (s.getVersion() >= VER(11)) {
 			for (i = ARRAYSIZE(_parts); i; --i, ++part) {
-				part->_instrument.saveOrLoad(ser);
+				part->_instrument.saveLoadWithSerializer(s);
 			}
 		} else {
 			for (i = ARRAYSIZE(_parts); i; --i, ++part)
@@ -435,10 +429,7 @@ int IMuseInternal::save_or_load(Serializer *ser, ScummEngine *scumm, bool fixAft
 	}
 
 	// VolumeFader has been replaced with the more generic ParameterFader.
-	// FIXME: replace this loop by something like
-	// if (loading && version <= 16)  ser->skip(XXX bytes);
-	for (i = 0; i < 8; ++i)
-		ser->saveLoadEntries(0, volumeFaderEntries);
+	s.skip(13 * 8, VER(8), VER(16));
 
 	// Normally, we have to fix up the data structures after loading a
 	// saved game. But there are cases where we don't. For instance, The
@@ -449,7 +440,7 @@ int IMuseInternal::save_or_load(Serializer *ser, ScummEngine *scumm, bool fixAft
 	// dummy iMUSE object, but since the resource is no longer recognizable
 	// to iMUSE, the fixup fails hard. So yes, this is a bit of a hack.
 
-	if (ser->isLoading() && fixAfterLoad) {
+	if (s.isLoading() && fixAfterLoad) {
 		// Load all sounds that we need
 		fix_players_after_load(scumm);
 		fix_parts_after_load();
@@ -460,8 +451,6 @@ int IMuseInternal::save_or_load(Serializer *ser, ScummEngine *scumm, bool fixAft
 		if (_midi_adlib)
 			reallocateMidiChannels(_midi_adlib);
 	}
-
-	return 0;
 }
 
 bool IMuseInternal::get_sound_active(int sound) const {
@@ -495,16 +484,17 @@ uint32 IMuseInternal::property(int prop, uint32 value) {
 	case IMuse::PROP_GS:
 		_enable_gs = (value > 0);
 
-		// If True Roland MT-32 is not selected, run in GM or GS mode.
-		// If it is selected, change the Roland GS synth to MT-32 mode.
-		if (_midi_native && !_native_mt32)
-			initGM(_midi_native);
-		else if (_midi_native && _native_mt32 && _enable_gs) {
-			_sc55 = true;
+		// GS Mode emulates MT-32 on a GS device, so _native_mt32 should always be true
+		if (_midi_native && _enable_gs) {
+			_native_mt32 = true;
 			initGM(_midi_native);
 		}
 		break;
 
+	case IMuse::PROP_AMIGA:
+		_isAmiga = (value > 0);
+		break;
+
 	case IMuse::PROP_LIMIT_PLAYERS:
 		if (value > 0 && value <= ARRAYSIZE(_players))
 			_player_limit = (int)value;
@@ -521,6 +511,9 @@ uint32 IMuseInternal::property(int prop, uint32 value) {
 	case IMuse::PROP_PC_SPEAKER:
 		_pcSpeaker = (value != 0);
 		break;
+
+	default:
+		break;
 	}
 
 	return 0;
@@ -614,7 +607,7 @@ bool IMuseInternal::startSound_internal(int sound, int offset) {
 	// playing, otherwise the music may stop when Sam and Max are thrown
 	// out of Bumpusville, because entering the mansion sets up a trigger
 	// for a sound that isn't necessarily playing. This is somewhat related
-	// to bug #780918.
+	// to bug #1090.
 
 	int i;
 	ImTrigger *trigger = _snm_triggers;
@@ -640,8 +633,8 @@ bool IMuseInternal::startSound_internal(int sound, int offset) {
 	// iMuse messiness while upgrading the iMuse engine, but it
 	// is apparently necessary to deal with fade-and-restart
 	// race conditions that were observed in MI2. Reference
-	// Bug #590511 and Patch #607175 (which was reversed to fix
-	// an FOA regression: Bug #622606).
+	// Bug #385 and Patch #8038 (which was reversed to fix
+	// an FOA regression: Bug #488).
 	Player *player = findActivePlayer(sound);
 	if (!player) {
 		ptr = findStartOfSound(sound, IMuseInternal::kMDhd);
@@ -679,6 +672,44 @@ bool IMuseInternal::startSound_internal(int sound, int offset) {
 	if (_game_id == GID_SAMNMAX && sound == 82 && getSoundStatus_internal(81, false))
 		ImClearTrigger(81, 1);
 
+	// Workaround for monkey2 bug #1410 / Scabb Island
+	//
+	// Tunes involved:
+	// 100 - Captain Dread's map
+	// 101 - Woodtick
+	// 107 - Map of Scabb Island
+	//
+	// If you go from Woodtick to the map of Scabb Island tune 107 is added as
+	// trigger on 101 and 101 moves to an outro and stop (triggering start
+	// of 107). Then at Captain Dread tune 107 is stopped and 100 is started.
+	//
+	// If you go quickly enough, the trigger occurs not at the Scabb Island
+	// map but at Captain Dread causing tune 107 not to be stopped.
+	// So we prevent starting 107 if 100 is already running.
+	if (_game_id == GID_MONKEY2 && (sound == 107) && (getSoundStatus_internal(100, true) == 1))
+		return false;
+
+	// In some cases 107 is running and doesn't get killed at Dread's map
+	if (_game_id == GID_MONKEY2 && (sound == 100) && (getSoundStatus_internal(107, true) == 1))
+		IMuseInternal::stopSound_internal(107);
+
+	// Workaround for monkey2 bug #1410 / Booty Island
+	//
+	// Tunes involved
+	// 100 - Captain Dread's map
+	// 113 - Guard Kiosk / Mardi Grass
+	// 115 - Map of Booty Island / Ville de la Booty
+	// 118 - Ville de la Booty
+	//
+	// When you enter the Guard Kiosk tune 113 is added as trigger on song
+	// 115. Then if you leave, 113 is stopped and 115 is started again.
+	// If you leave quickly enough, the trigger occurs on the map and tune
+	// 113 will not stop.
+	// We kill 113 on entry of one of the other locations (Captain Dread
+	// or Ville de la Booty) because tune 115 is not always started.
+	if (_game_id == GID_MONKEY2 && (sound == 100 || sound == 115 || sound == 118) && (getSoundStatus_internal(113, true) == 1))
+		IMuseInternal::stopSound_internal(113);
+
 	player->clear();
 	player->setOffsetNote(offset);
 	return player->startSound(sound, driver);
@@ -918,7 +949,7 @@ int32 IMuseInternal::doCommand_internal(int numargs, int a[]) {
 			player->setDetune(a[2]);
 			return 0;
 		case 6:
-			// WORKAROUND for bug #1324106. When playing the
+			// WORKAROUND for bug #2242. When playing the
 			// "flourishes" as Rapp's body appears from his ashes,
 			// MI2 sets up triggers to pause the music, in case the
 			// animation plays too slowly, and then the music is
@@ -1224,15 +1255,12 @@ int32 IMuseInternal::ImSetTrigger(int sound, int id, int a, int b, int c, int d,
 			break;
 		// We used to only compare 'id' and 'sound' here, but at least
 		// at the Dino Bungie Memorial that causes the music to stop
-		// after getting the T-Rex tooth. See bug #888161.
+		// after getting the T-Rex tooth. See bug #1429.
 		if (trig->id == id && trig->sound == sound && trig->command[0] == a)
 			break;
 
-		uint16 diff;
-		if (trig->expire <= _snm_trigger_index)
-			diff = _snm_trigger_index - trig->expire;
-		else
-			diff = 0x10000 - trig->expire + _snm_trigger_index;
+		// The wraparound if trig->expire > _snm_trigger_index is intentional
+		uint16 diff = _snm_trigger_index - trig->expire;
 
 		if (!oldest_ptr || oldest_trigger < diff) {
 			oldest_ptr = trig;
@@ -1263,7 +1291,7 @@ int32 IMuseInternal::ImSetTrigger(int sound, int id, int a, int b, int c, int d,
 	// This fixes some carnival music problems.
 	// NOTE: We ONLY do this if the sound that will trigger the command is actually
 	// playing. Otherwise, there's a problem when exiting and re-entering the
-	// Bumpusville mansion. Ref Bug #780918.
+	// Bumpusville mansion. Ref Bug #1090.
 	if (trig->command[0] == 8 && getSoundStatus_internal(trig->command[1], true) && getSoundStatus_internal(sound, true))
 		stopSound_internal(trig->command[1]);
 	return 0;
@@ -1452,8 +1480,6 @@ void IMuseInternal::initMidiDriver(TimerCallbackInfo *info) {
 
 void IMuseInternal::initMT32(MidiDriver *midi) {
 	byte buffer[52];
-	char info[256] = "ScummVM ";
-	int len;
 
 	// Reset the MT-32
 	midi->sysEx((const byte *) "\x41\x10\x16\x12\x7f\x00\x00\x01\x00", 9);
@@ -1469,15 +1495,16 @@ void IMuseInternal::initMT32(MidiDriver *midi) {
 	_system->delayMillis(250);
 
 	// Compute version string (truncated to 20 chars max.)
-	strcat(info, gScummVMVersion);
-	len = strlen(info);
+	Common::String infoStr = "ScummVM ";
+	infoStr += gScummVMVersion;
+	int len = infoStr.size();
 	if (len > 20)
 		len = 20;
 
 	// Display a welcome message on MT-32 displays.
 	memcpy(&buffer[0], "\x41\x10\x16\x12\x20\x00\x00", 7);
 	memcpy(&buffer[7], "                    ", 20);
-	memcpy(buffer + 7 + (20 - len) / 2, info, len);
+	memcpy(buffer + 7 + (20 - len) / 2, infoStr.c_str(), len);
 	byte checksum = 0;
 	for (int i = 4; i < 27; ++i)
 		checksum -= buffer[i];
@@ -1487,7 +1514,7 @@ void IMuseInternal::initMT32(MidiDriver *midi) {
 }
 
 void IMuseInternal::initGM(MidiDriver *midi) {
-	byte buffer[11];
+	byte buffer[12];
 	int i;
 
 	// General MIDI System On message
@@ -1499,7 +1526,7 @@ void IMuseInternal::initGM(MidiDriver *midi) {
 
 	if (_enable_gs) {
 		// All GS devices recognize the GS Reset command,
-		// even with Roland's ID. It is impractical to
+		// even using Roland's ID. It is impractical to
 		// support other manufacturers' devices for
 		// further GS settings, as there are limitless
 		// numbers of them out there that would each
@@ -1514,29 +1541,27 @@ void IMuseInternal::initGM(MidiDriver *midi) {
 		debug(2, "GS SysEx: GS Reset");
 		_system->delayMillis(200);
 
-		if (_sc55) {
-			// This mode is for GS devices that support an MT-32-compatible
-			// Map, such as the Roland Sound Canvas line of modules. It
-			// will allow them to work with True MT-32 mode, but will
-			// obviously still ignore MT-32 SysEx (and thus custom
-			// instruments).
+		// Set global Master Tune to 442.0kHz, as on the MT-32
+		memcpy(&buffer[4], "\x40\x00\x00\x00\x04\x04\x0F\x29", 8);
+		midi->sysEx(buffer, 12);
+		debug(2, "GS SysEx: Master Tune set to 442.0kHz");
 
-			// Set Channels 1-16 to SC-55 Map, then CM-64/32L Variation
-			for (i = 0; i < 16; ++i) {
-				midi->send((127 << 16) | (0  << 8) | (0xB0 | i));
-				midi->send((1   << 16) | (32 << 8) | (0xB0 | i));
-				midi->send((0   << 16) | (0  << 8) | (0xC0 | i));
-			}
-			debug(2, "GS Program Change: CM-64/32L Map Selected");
-
-			// Set Percussion Channel to SC-55 Map (CC#32, 01H), then
-			// Switch Drum Map to CM-64/32L (MT-32 Compatible Drums)
-			midi->getPercussionChannel()->controlChange(0, 0);
-			midi->getPercussionChannel()->controlChange(32, 1);
-			midi->send(127 << 8 | 0xC0 | 9);
-			debug(2, "GS Program Change: Drum Map is CM-64/32L");
+		// Note: All Roland GS devices support CM-64/32L maps
 
+		// Set Channels 1-16 to SC-55 Map, then CM-64/32L Variation
+		for (i = 0; i < 16; ++i) {
+			midi->send((127 << 16) | (0  << 8) | (0xB0 | i));
+			midi->send((1   << 16) | (32 << 8) | (0xB0 | i));
+			midi->send((0   << 16) | (0  << 8) | (0xC0 | i));
 		}
+		debug(2, "GS Program Change: CM-64/32L Map Selected");
+
+		// Set Percussion Channel to SC-55 Map (CC#32, 01H), then
+		// Switch Drum Map to CM-64/32L (MT-32 Compatible Drums)
+		midi->getPercussionChannel()->controlChange(0, 0);
+		midi->getPercussionChannel()->controlChange(32, 1);
+		midi->send(127 << 8 | 0xC0 | 9);
+		debug(2, "GS Program Change: Drum Map is CM-64/32L");
 
 		// Set Master Chorus to 0. The MT-32 has no chorus capability.
 		memcpy(&buffer[4], "\x40\x01\x3A\x00\x05", 5);
@@ -1546,7 +1571,7 @@ void IMuseInternal::initGM(MidiDriver *midi) {
 		// Set Channels 1-16 Reverb to 64, which is the
 		// equivalent of MT-32 default Reverb Level 5
 		for (i = 0; i < 16; ++i)
-			midi->send((64   << 16) | (91 << 8) | (0xB0 | i));
+			midi->send((64 << 16) | (91 << 8) | (0xB0 | i));
 		debug(2, "GM Controller 91 Change: Channels 1-16 Reverb Level is 64");
 
 		// Set Channels 1-16 Pitch Bend Sensitivity to
